@@ -1,14 +1,21 @@
+from functools import reduce
 import math
 import multiprocessing as mp
 import os
 import sys
+import time
+import webbrowser
 
 import click
-from cloudvolume import CloudVolume
+from cloudvolume import CloudVolume, Bbox
+from cloudvolume.lib import max2
+from cloudfiles import CloudFiles
+import cloudfiles.paths
 import numpy as np
 from taskqueue import TaskQueue
 from taskqueue.lib import toabs
 from taskqueue.paths import get_protocol
+from tqdm import tqdm
 
 from igneous import task_creation as tc
 from igneous import downsample_scales
@@ -49,7 +56,7 @@ class Tuple2(click.ParamType):
 
 @click.group()
 @click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
-@click.version_option(version="0.8.1")
+@click.version_option(version="0.9.2")
 @click.pass_context
 def main(ctx, parallel):
   """
@@ -101,9 +108,9 @@ def license():
 @click.option('--bg-color', default=0, help="Determines which color is regarded as background. Default: 0")
 @click.option('--sharded', is_flag=True, default=False, help="Generate sharded downsamples which reduces the number of files.")
 @click.option('--memory', default=3.5e9, type=int, help="(sharded only) Task memory limit in bytes. Task shape will be chosen to fit and maximize downsamples.", show_default=True)
-@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for downsampling. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024.", show_default=True)
-@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for downsampling. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024", show_default=True)
-@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for downsampling. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1", show_default=True)
+@click.option('--xrange', type=Tuple2(), default=None, help="If specified, set x-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024.", show_default=True)
+@click.option('--yrange', type=Tuple2(), default=None, help="If specified, set y-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1024", show_default=True)
+@click.option('--zrange', type=Tuple2(), default=None, help="If specified, set z-bounds for downsampling in terms of selected mip. By default the whole dataset is selected. The bounds must be chunk aligned to the task size (maybe mysterious... use igneous design to investigate). e.g. 0,1", show_default=True)
 @click.pass_context
 def downsample(
   ctx, path, queue, mip, fill_missing, 
@@ -126,6 +133,8 @@ def downsample(
   current top mip level of the pyramid. This builds it even taller
   (referred to as "superdownsampling").
   """
+  path = cloudfiles.paths.normalize(path)
+
   if cseg and compresso:
     print("igneous: must choose one of --cseg or --compresso")
     return
@@ -163,7 +172,7 @@ def downsample(
       path, mip=mip, fill_missing=fill_missing, 
       sparse=sparse, chunk_size=chunk_size,
       encoding=encoding, memory_target=memory,
-      factor=factor, bounds=bounds
+      factor=factor, bounds=bounds, bounds_mip=mip
     )
   else:
     tasks = tc.create_downsampling_tasks(
@@ -173,7 +182,8 @@ def downsample(
       delete_black_uploads=delete_bg, 
       background_color=bg_color, 
       compress=compress,
-      factor=factor, bounds=bounds
+      factor=factor, bounds=bounds,
+      bounds_mip=mip
     )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -202,6 +212,7 @@ def downsample(
 @click.option('--sharded', is_flag=True, default=False, help="Generate a sharded dataset which reduces the number of files. Downsamples are not generated.")
 @click.option('--dest-voxel-offset', type=Tuple3(), default=None, help="Set the voxel offset for this mip level.")
 @click.option('--clean-info', is_flag=True, default=False, help="Scrub info file of mesh and skeleton fields.", show_default=True)
+@click.option('--no-src-update', is_flag=True, default=False, help="Don't update the source provenance file with the transfer metadata.", show_default=True)
 @click.pass_context
 def xfer(
 	ctx, src, dest, queue, translate, 
@@ -209,7 +220,7 @@ def xfer(
   memory, max_mips, shape, sparse, 
   cseg, compresso, chunk_size, compress, 
   volumetric, delete_bg, bg_color, sharded,
-  dest_voxel_offset, clean_info
+  dest_voxel_offset, clean_info, no_src_update
 ):
   """
   Transfer an image layer to another location.
@@ -229,6 +240,9 @@ def xfer(
   Use the --memory flag to automatically compute the
   a reasonable task shape based on your memory limits.
   """
+  src = cloudfiles.paths.normalize(src)
+  dest = cloudfiles.paths.normalize(dest)
+
   if cseg and compresso:
     print("igneous: must choose one of --cseg or --compresso")
     sys.exit()
@@ -261,7 +275,8 @@ def xfer(
       mip=mip, shape=shape, encoding=encoding, skip_downsamples=(not downsample),
       delete_black_uploads=delete_bg, background_color=bg_color,
       compress=compress, factor=factor, sparse=sparse,
-      memory_target=memory, max_mips=max_mips, clean_info=clean_info
+      memory_target=memory, max_mips=max_mips, 
+      clean_info=clean_info, no_src_update=no_src_update
     )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -270,12 +285,17 @@ def xfer(
 
 @main.command()
 @click.argument("queue", type=str)
-@click.option('--aws-region', default=SQS_REGION_NAME, help=f"AWS region in which the SQS queue resides. Default: {SQS_REGION_NAME}")
-@click.option('--lease-sec', default=LEASE_SECONDS, help=f"Seconds to lease a task for. Default: {LEASE_SECONDS}", type=int)
-@click.option('--tally/--no-tally', is_flag=True, default=True, help="Tally completed fq tasks. Does not apply to SQS.")
+@click.option('--aws-region', default=SQS_REGION_NAME, help=f"AWS region in which the SQS queue resides.", show_default=True)
+@click.option('--lease-sec', default=LEASE_SECONDS, help=f"Seconds to lease a task for.", type=int, show_default=True)
+@click.option('--tally/--no-tally', is_flag=True, default=True, help="Tally completed fq tasks. Does not apply to SQS.", show_default=True)
 @click.option('--min-sec', default=-1, help='Execute for at least this many seconds and quit after the last task finishes. Special values: (0) Run at most a single task. (-1) Loop forever (default).', type=float)
+@click.option('-x', '--exit-on-empty', is_flag=True, default=False, help="Exit immediately when the queue is empty. Not reliable for SQS as measurements are approximate.", show_default=True)
 @click.pass_context
-def execute(ctx, queue, aws_region, lease_sec, tally, min_sec):
+def execute(
+  ctx, queue, aws_region,
+  lease_sec, tally, min_sec,
+  exit_on_empty
+):
   """Execute igneous tasks from a queue.
 
   The queue must be an AWS SQS queue or a FileQueue directory. 
@@ -285,7 +305,7 @@ def execute(ctx, queue, aws_region, lease_sec, tally, min_sec):
   See https://github.com/seung-lab/python-task-queue
   """
   parallel = int(ctx.obj.get("parallel", 1))
-  args = (queue, aws_region, lease_sec, tally, min_sec)
+  args = (queue, aws_region, lease_sec, tally, min_sec, exit_on_empty)
 
   if parallel == 1:
     execute_helper(*args)
@@ -302,10 +322,16 @@ def execute(ctx, queue, aws_region, lease_sec, tally, min_sec):
     pool.terminate()
     pool.join()
 
-def execute_helper(queue, aws_region, lease_sec, tally, min_sec):
+def execute_helper(
+  queue, aws_region, lease_sec, 
+  tally, min_sec, exit_on_empty
+):
   tq = TaskQueue(normalize_path(queue), region_name=aws_region)
 
-  def stop_after_elapsed_time(elapsed_time):
+  def stop_after_elapsed_time(tries, elapsed_time):
+    if exit_on_empty and tq.is_empty():
+      return True
+
     if min_sec < 0:
       return False
     return min_sec < elapsed_time
@@ -336,21 +362,22 @@ def meshgroup():
 @meshgroup.command("forge")
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
-@click.option('--mip', default=0, help="Perform meshing using this level of the image pyramid. Default: 0")
-@click.option('--shape', type=Tuple3(), default=(448, 448, 448), help="Set the task shape in voxels. Default: 448,448,448")
-@click.option('--simplify/--skip-simplify', is_flag=True, default=True, help="Enable mesh simplification. Default: True")
+@click.option('--mip', default=0, help="Perform meshing using this level of the image pyramid.", show_default=True)
+@click.option('--shape', type=Tuple3(), default=(448, 448, 448), help="Set the task shape in voxels.", show_default=True)
+@click.option('--simplify/--skip-simplify', is_flag=True, default=True, help="Enable mesh simplification.", show_default=True)
 @click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
-@click.option('--max-error', default=40, help="Maximum simplification error in physical units. Default: 40 nm")
+@click.option('--max-error', default=40, help="Maximum simplification error in physical units.", show_default=True)
 @click.option('--dust-threshold', default=None, help="Skip meshing objects smaller than this number of voxels within a cutout. No default limit. Typical value: 1000.", type=int)
 @click.option('--dir', default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
-@click.option('--compress', default=None, help="Set the image compression scheme. Options: 'gzip', 'br'")
-@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.")
+@click.option('--compress', default="gzip", help="Set the image compression scheme. Options: 'none', 'gzip', 'br'", show_default=True)
+@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.", show_default=True)
+@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing mesh fragments.", show_default=True)
 @click.pass_context
 def mesh_forge(
   ctx, path, queue, mip, shape, 
   simplify, fill_missing, max_error, 
   dust_threshold, dir, compress, 
-  spatial_index
+  spatial_index, sharded
 ):
   """
   (1) Synthesize meshes from segmentation cutouts.
@@ -367,13 +394,17 @@ def mesh_forge(
 
   Sharded format not currently supports. Coming soon.
   """
+  path = cloudfiles.paths.normalize(path)
+  if compress.lower() == "none":
+    compress = False
+
   tasks = tc.create_meshing_tasks(
     path, mip, shape, 
     simplification=simplify, max_simplification_error=max_error,
     mesh_dir=dir, cdn_cache=False, dust_threshold=dust_threshold,
     object_ids=None, progress=False, fill_missing=fill_missing,
     encoding='precomputed', spatial_index=spatial_index, 
-    sharded=False, compress=compress
+    sharded=sharded, compress=compress
   )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -395,6 +426,7 @@ def mesh_merge(ctx, path, queue, magnitude, dir):
   a list of fragment files and uploading a "mesh manifest"
   file that is an index for locating the fragments.
   """
+  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_mesh_manifest_tasks(
     path, magnitude=magnitude, mesh_dir=dir
   )
@@ -403,6 +435,99 @@ def mesh_merge(ctx, path, queue, magnitude, dir):
   tq = TaskQueue(normalize_path(queue))
   tq.insert(tasks, parallel=parallel)
 
+@meshgroup.command("merge-sharded")
+@click.argument("path")
+@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
+@click.option('--vqb', default=16, help="Vertex quantization bits. Can be 10 or 16.", type=int, show_default=True)
+@click.option('--compress-level', default=1, help="Draco compression level.", type=int, show_default=True)
+@click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
+@click.option('--minishard-index-bytes', default=2**15, help="Size in bytes to make the minishard index.", type=int, show_default=True)
+@click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw.", show_default=True)
+@click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
+@click.pass_context
+def mesh_sharded_merge(
+  ctx, path, queue, 
+  vqb, compress_level,
+  shard_index_bytes, minishard_index_bytes,
+  minishard_index_encoding, spatial_index_db
+):
+  """
+  (2) Postprocess fragments into finished sharded multires meshes.
+
+  Only use this command if you used the --sharded flag
+  during the forging step. Some reasonable defaults
+  are selected for a dataset with a few million labels,
+  but for smaller or larger datasets they may not be
+  appropriate.
+
+  The shard and minishard index default sizes are set to
+  accomodate efficient access for a 100 Mbps connection.
+  """
+  path = cloudfiles.paths.normalize(path)
+  tasks = tc.create_sharded_multires_mesh_tasks(
+    path, 
+    draco_compression_level=compress_level,
+    vertex_quantization_bits=vqb,
+    shard_index_bytes=shard_index_bytes,
+    minishard_index_bytes=minishard_index_bytes,
+    minishard_index_encoding=minishard_index_encoding,
+    spatial_index_db=spatial_index_db,
+  )
+
+  parallel = int(ctx.obj.get("parallel", 1))
+  tq = TaskQueue(normalize_path(queue))
+  tq.insert(tasks, parallel=parallel)
+
+
+@meshgroup.group("spatial-index")
+def spatialindexgroup():
+  """
+  (subgroup) Create or download mesh spatial indices.
+  """
+  pass
+
+@spatialindexgroup.command("create")
+@click.argument("path")
+@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
+@click.option('--shape', default="448,448,448", type=Tuple3(), help="Shape in voxels of each indexing task.", show_default=True)
+@click.option('--mip', default=0, help="Perform indexing using this level of the image pyramid.", show_default=True)
+@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
+@click.pass_context
+def mesh_spatial_index_create(ctx, path, queue, shape, mip, fill_missing):
+  """
+  Create a spatial index on a pre-existing mesh.
+
+  Sometimes datasets were meshes without a
+  spatial index or need it to be updated.
+  This function provides a more efficient
+  way to accomplish that than remeshing.
+  """
+  path = cloudfiles.paths.normalize(path)
+  tasks = tc.create_spatial_index_mesh_tasks(
+    cloudpath=path,
+    shape=shape,
+    mip=mip, 
+    fill_missing=fill_missing,
+  )
+
+  parallel = int(ctx.obj.get("parallel", 1))
+  tq = TaskQueue(normalize_path(queue))
+  tq.insert(tasks, parallel=parallel)
+
+@spatialindexgroup.command("db")
+@click.argument("path")
+@click.argument("database")
+@click.option('--progress', is_flag=True, default=False, help="Show progress bars.", show_default=True)
+def mesh_spatial_index_download(path, database, progress):
+  """
+  Download the mesh spatial index into a database.
+
+  sqlite paths: sqlite://filename.db (prefix optional)
+  mysql paths: mysql://{user}:{pwd}@{host}/{database}
+  """
+  path = cloudfiles.paths.normalize(path)
+  cv = CloudVolume(path)
+  cv.mesh.spatial_index.to_sql(database, progress=progress)
 
 @main.group("skeleton")
 def skeletongroup():
@@ -420,23 +545,23 @@ def skeletongroup():
 @skeletongroup.command("forge")
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
-@click.option('--mip', default=0, help="Perform skeletonizing using this level of the image pyramid. Default: 0")
-@click.option('--shape', type=Tuple3(), default=(512, 512, 512), help="Set the task shape in voxels. Default: 512,512,512")
-@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
+@click.option('--mip', default=0, help="Perform skeletonizing using this level of the image pyramid.", show_default=True)
+@click.option('--shape', type=Tuple3(), default=(512, 512, 512), help="Set the task shape in voxels.", show_default=True)
+@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.", show_default=True)
 @click.option('--fix-branching', is_flag=True, default=True, help="Trades speed for quality of branching at forks. Default: True")
-@click.option('--fix-borders', is_flag=True, default=True, help="Allows trivial merging of single voxel overlap tasks. Only switch off for datasets that fit in a single task. Default: True")
-@click.option('--fix-avocados', is_flag=True, default=False, help="Fixes somata where nuclei and cytoplasm have separate segmentations. Default: False")
-@click.option('--fill-holes', is_flag=True, default=False, help="Preprocess each cutout to eliminate background holes and holes caused by entirely contained inclusions. Warning: May remove labels that are considered inclusions. Default: False")
-@click.option('--dust-threshold', default=1000, help="Skip skeletonizing objects smaller than this number of voxels within a cutout. Default: 1000.", type=int)
-@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.")
-@click.option('--scale', default=4, help="Multiplies invalidation radius by distance from boundary.", type=float)
-@click.option('--const', default=10, help="Adds constant amount to invalidation radius in physical units.", type=float)
-@click.option('--soma-detect', default=1100, help="Consider objects with peak distances to boundary larger than this soma candidates. Physical units. Default: 1100 nm", type=float)
-@click.option('--soma-accept', default=3500, help="Accept soma candidates over this threshold and perform a one-time spherical invalidation around their peak value. Physical units. Default: 3500 nm", type=float)
-@click.option('--soma-scale', default=1.0, help="Scale factor for soma invalidation.", type=float)
-@click.option('--soma-const', default=300, help="Const factor for soma invalidation.", type=float)
+@click.option('--fix-borders', is_flag=True, default=True, help="Allows trivial merging of single voxel overlap tasks. Only switch off for datasets that fit in a single task.", show_default=True)
+@click.option('--fix-avocados', is_flag=True, default=False, help="Fixes somata where nuclei and cytoplasm have separate segmentations.", show_default=True)
+@click.option('--fill-holes', is_flag=True, default=False, help="Preprocess each cutout to eliminate background holes and holes caused by entirely contained inclusions. Warning: May remove labels that are considered inclusions.", show_default=True)
+@click.option('--dust-threshold', default=1000, help="Skip skeletonizing objects smaller than this number of voxels within a cutout.", type=int, show_default=True)
+@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.", show_default=True)
+@click.option('--scale', default=4, help="Multiplies invalidation radius by distance from boundary.", type=float, show_default=True)
+@click.option('--const', default=10, help="Adds constant amount to invalidation radius in physical units.", type=float, show_default=True)
+@click.option('--soma-detect', default=1100, help="Consider objects with peak distances to boundary larger than this soma candidates. Physical units.", type=float, show_default=True)
+@click.option('--soma-accept', default=3500, help="Accept soma candidates over this threshold and perform a one-time spherical invalidation around their peak value. Physical units.", type=float, show_default=True)
+@click.option('--soma-scale', default=1.0, help="Scale factor for soma invalidation.", type=float, show_default=True)
+@click.option('--soma-const', default=300, help="Const factor for soma invalidation.", type=float, show_default=True)
 @click.option('--max-paths', default=None, help="Abort skeletonizing an object after this many paths have been traced.", type=float)
-@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing skeleton fragments.")
+@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing skeleton fragments.", show_default=True)
 @click.pass_context
 def skeleton_forge(
   ctx, path, queue, mip, shape, 
@@ -465,6 +590,7 @@ def skeleton_forge(
 
   - https://github.com/seung-lab/kimimaro/wiki/The-Economics:-Skeletons-for-the-People
   """
+  path = cloudfiles.paths.normalize(path)
   teasar_params = {
     'scale': scale,
     'const': const, # physical units
@@ -510,6 +636,7 @@ def skeleton_merge(
   """
   (2) Postprocess fragments into finished skeletons.
   """
+  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_unsharded_skeleton_merge_tasks(
     path, 
     magnitude=magnitude, 
@@ -526,21 +653,22 @@ def skeleton_merge(
 @skeletongroup.command("merge-sharded")
 @click.argument("path")
 @click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
-@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length. Default: 1000 nm", type=float)
+@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length.", type=float, show_default=True)
 @click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
 @click.option('--tick-threshold', default=0, help="Remove small \"ticks\", or branches from the main skeleton one at a time from smallest to largest. Branches larger than this are preserved. Default: no elimination", type=float)
-@click.option('--preshift-bits', default=3, help="Shift LSB to try to increase number of hash collisions.", type=int)
-@click.option('--minishard-bits', default=10, help="2^bits number of bays holding variable numbers of labels per shard.", type=int)
-@click.option('--shard-bits', default=2, help="2^bits number of shard files to generate.", type=int)
+@click.option('--shard-index-bytes', default=2**13, help="Size in bytes to make the shard index.", type=int, show_default=True)
+@click.option('--minishard-index-bytes', default=2**15, help="Size in bytes to make the minishard index.", type=int, show_default=True)
 @click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw. Default: gzip")
 @click.option('--data-encoding', default="gzip", help="Shard data can be compressed. gzip or raw. Default: gzip")
+@click.option('--spatial-index-db', default=None, help="CloudVolume generated SQL database for spatial index.", show_default=True)
 @click.pass_context
 def skeleton_sharded_merge(
   ctx, path, queue, 
   min_cable_length, max_cable_length, 
   tick_threshold, 
-  preshift_bits, minishard_bits, shard_bits,
-  minishard_index_encoding, data_encoding
+  shard_index_bytes, minishard_index_bytes,
+  minishard_index_encoding, data_encoding,
+  spatial_index_db
 ):
   """
   (2) Postprocess fragments into finished skeletons.
@@ -550,17 +678,20 @@ def skeleton_sharded_merge(
   are selected for a dataset with a few million labels,
   but for smaller or larger datasets they may not be
   appropriate.
+
+  The shard and minishard index default sizes are set to
+  accomodate efficient access for a 100 Mbps connection.
   """
   tasks = tc.create_sharded_skeleton_merge_tasks(
     path, 
     dust_threshold=min_cable_length,
     max_cable_length=max_cable_length,
     tick_threshold=tick_threshold,
-    preshift_bits=preshift_bits, 
-    minishard_bits=minishard_bits, 
-    shard_bits=shard_bits,
+    shard_index_bytes=shard_index_bytes,
+    minishard_index_bytes=minishard_index_bytes,
     minishard_index_encoding=minishard_index_encoding, 
     data_encoding=data_encoding,
+    spatial_index_db=spatial_index_db,
   )
 
   parallel = int(ctx.obj.get("parallel", 1))
@@ -589,6 +720,7 @@ def delete_images(
   """
   Delete the image layer of a dataset.
   """
+  path = cloudfiles.paths.normalize(path)
   tasks = tc.create_deletion_tasks(
     path, mip, num_mips=num_mips, shape=shape
   )
@@ -603,6 +735,29 @@ def designgroup():
   """
   pass
 
+@designgroup.command("bounds")
+@click.argument("path")
+@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
+def dsbounds(path, mip):
+  """
+  Detects the volume bounds and chunk size for
+  an unsharded image volume. Useful when there
+  is a corrupted info file.
+  """
+  path = cloudfiles.paths.normalize(path)
+  cv = CloudVolume(path, mip=mip)
+  cf = CloudFiles(path)
+
+  bboxes = []
+  for filename in tqdm(cf.list(prefix=cv.key), desc="Computing Bounds"):
+    bboxes.append( Bbox.from_filename(filename) )
+
+  bounds = Bbox.expand(*bboxes)
+  chunk_size = list(reduce(max2, map(lambda bbox: bbox.size3(), bboxes)))
+
+  print(f"Bounds: {bounds}")
+  print(f"Volume: {list(bounds.size3())}")
+  print(f"Chunks: {chunk_size}")
 
 @designgroup.command("ds-memory")
 @click.argument("path")
@@ -616,6 +771,7 @@ def dsmemory(path, memory_bytes, mip, factor, verbose, max_mips):
   Compute the task shape that maximizes the number of
   downsamples for a given amount of memory.
   """ 
+  path = cloudfiles.paths.normalize(path)
   cv = CloudVolume(path, mip=mip)
 
   data_width = np.dtype(cv.dtype).itemsize
@@ -663,6 +819,7 @@ def dsshape(path, shape, mip, factor):
   Compute the approximate memory usage for a
   given downsample task shape.
   """ 
+  path = cloudfiles.paths.normalize(path)
   cv = CloudVolume(path, mip=mip)
   data_width = np.dtype(cv.dtype).itemsize
   memory_bytes = memory_used(data_width, shape, factor)  
@@ -681,3 +838,21 @@ def memory_used(data_width, shape, factor):
     memory_bytes *= constant / (constant - 1)
 
   return memory_bytes
+
+@main.command("view")
+@click.argument("path")
+@click.option('--browser/--no-browser', default=True, is_flag=True, help="Open the dataset in the system's default web browser.")
+@click.option('--port', default=1337, help="localhost server port for the file server.", show_default=True)
+@click.option('--ng', default="https://neuroglancer-demo.appspot.com/", help="Alternative Neuroglancer webpage to use.", show_default=True)
+def view(path, browser, port, ng):
+  """
+  Open an on-disk dataset for viewing in neuroglancer.
+  """
+  # later improvements: 
+  #   could use local neuroglancer
+  #   modify the url to autopopulate params to avoid a click
+  path = cloudfiles.paths.normalize(path)
+  url = f"{ng}#!%7B%22layers%22:%5B%7B%22type%22:%22new%22%2C%22source%22:%22precomputed://http://localhost:{port}%22%2C%22tab%22:%22source%22%2C%22name%22:%22localhost:{port}%22%7D%5D%2C%22selectedLayer%22:%7B%22visible%22:true%2C%22layer%22:%22localhost:{port}%22%7D%2C%22layout%22:%224panel%22%7D"
+  if browser:
+    webbrowser.open(url, new=2)
+  CloudVolume(path).viewer(port=port)
